from .. import Sampler
import numpy as np
from ...runtime import pipeline
import sys
import os
import io
import glob
import pickle
from collections import defaultdict

matter_power_section_names = [
    'dark_matter_power',
    'baryon_power',
    'photon_power',
    'massless_neutrino_power',
    'massive_neutrino_power',
    'matter_power',
    'cdm_baryon_power',
    'matter_de_power',
    'weyl_curvature_power',
    'cdm_velocity_power',
    'baryon_velocity_power',
    'baryon_cdm_relative_velocity_power',
]

class CosmoPowerSampler(Sampler):
    parallel_output = False
    needs_output = False

    def load_pickles(self, params, filenames):
        """
        Load the pickled files generated by the Latin Hypercube sampler which saved the datablock information for each parameter evaluation
        and extract the relevant information for training/testing the CosmoPower emulator, in a format that can be used by the cosmopower package.
        Currently only supports CAMB outputs.

        Parameters:
        - params: list, list of parameter objects used in the pipeline
        - filenames: list, list of filenames to load
        """
        index_list = []
        params_dict = defaultdict(list)
        out_dict = defaultdict(lambda: defaultdict(list))
        runs = set()

        for filename in filenames:
            index_list.append(int(os.path.basename(filename).split('_')[-1].split('.')[0]))
            with open(filename, 'rb') as f:
                data = pickle.load(f)
        
            # This could be simplified and generalised for all inputs, by having the same function as block.from_pickle, 
            # but leaving now as it is to extract only the relevant quantities
            for param in params:
                p = param.name
                # Extend the option where the parameters can be and which spectra we have!
                for key in ['cosmological_parameters', 'halo_model_parameters', 'redshift_as_parameter']:
                    if p in data.get(key):
                        params_dict[p].append(data[key][p])

            suffixes = ['lin', 'nl']
            all_keys = [f'{key}_{suffix}' for key in matter_power_section_names for suffix in suffixes]
            all_keys.append('cmb_cl')
            for key in all_keys:
                if data.get(key):
                    runs.add(key)
                    if key != 'cmb_cl':
                        out_dict[key]['k_h'].append(np.squeeze(data[key]['k_h']).tolist())
                        out_dict[key]['p_k'].append(np.squeeze(data[key]['p_k']).tolist())
                    elif key == 'cmb_cl':
                        for cl in ['ell', 'tt', 'ee', 'bb', 'te']:
                            out_dict[key][cl].append(np.squeeze(data[key][cl]).tolist())
                        for cl in ['pp', 'pt', 'pe']:
                            if cl in data[key]:
                                out_dict[key][cl].append(np.squeeze(data[key][cl]).tolist())

        return index_list, list(runs), params_dict, dict(out_dict)
    
    def run_training_pk(self, tf, cosmopower_NN, cosmopower_PCAplusNN, run, params_dict_train, runs_dict, mean_run):
        """
        Train the neural network for the matter power spectrum emulation.

        Parameters:
        - run: str, the name of the power spectrum to train on (e.g., 'p_k')
        - params_dict_train: dict, dictionary of training parameters
        - runs_dict: dict, dictionary of training power spectra
        - mean_run: object, the mean run from the pipeline to use as reference prediction   
        """
        reference_prediction = np.squeeze(mean_run.block[run, 'p_k'])
        # Save reference prediction to file
        with open(f'{self.save_dir}_{run}_reference.pkl', 'wb') as f:
            pickle.dump(reference_prediction, f)
        
        # Initiate and train the neural network
        cp_nn = cosmopower_NN(parameters=list(params_dict_train.keys()),
                            modes=runs_dict[run]['k_h'][0],
                            n_hidden=[1048, 1048, 1048, 1048],
                            verbose=True)
        
        training_features = np.log10(runs_dict[run]['p_k']) - np.log10(reference_prediction)
        with tf.device(self.device):
            # The training choices are for now hardcoded. These showed to perform well for any CAMB outputs.
            cp_nn.train(training_parameters=params_dict_train,
                        training_features=training_features,
                        filename_saved_model=f'{self.save_dir}_{run}',
                        validation_split=0.1,
                        learning_rates=[1e-2, 1e-3, 1e-4, 1e-5, 1e-6, 1e-7],
                        batch_sizes=[1000, 1000, 1000, 1000, 1000, 1000],
                        gradient_accumulation_steps=[1, 1, 1, 1, 1, 1],
                        patience_values=[100, 100, 100, 100, 100, 100],
                        max_epochs=[1000, 1000, 1000, 1000, 1000, 1000])
        
    def run_testing_pk(self, tf, cosmopower_NN, cosmopower_PCAplusNN, run, params_dict_test, test_dict, mean_run):
        """
        Test the neural network for the matter power spectrum emulation and plot the results.

        Parameters:
        - run: str, the name of the power spectrum to test on (e.g., 'p_k')
        - params_dict_test: dict, dictionary of testing parameters
        - test_dict: dict, dictionary of testing power spectra
        - mean_run: object, the mean run from the pipeline to use as reference prediction
        """
        import matplotlib.pyplot as plt
        reference_prediction = np.squeeze(mean_run.block[run, 'p_k'])
        # Enter testing phase
        cp_nn_test = cosmopower_NN(restore=True, restore_filename=f'{self.save_dir}_{run}')
        emulated_test_spectra = 10.0**(cp_nn_test.predictions_np(params_dict_test) + np.log10(reference_prediction))
        diff = np.abs((emulated_test_spectra - test_dict[run]['p_k']) / test_dict[run]['p_k'])
        # Define percentiles to calculate
        percentile_values = [68, 95, 99, 99.9]
        percentiles = np.array([np.percentile(diff, p, axis=0) for p in percentile_values])
        # Define colors and labels for plotting
        plot_settings = [
            {'color': 'salmon', 'label': '99%', 'alpha': 0.8},
            {'color': 'red', 'label': '95%', 'alpha': 0.7},
            {'color': 'darkred', 'label': '68%', 'alpha': 1}
        ]
        plt.figure(figsize=(12, 9))
        for i, settings in enumerate(plot_settings[::-1]):
            plt.fill_between(test_dict[run]['k_h'][0], 0, percentiles[i], **settings)
        plt.xscale('log')
        plt.legend(frameon=False, fontsize=30, loc='upper left')
        plt.ylabel(r'$\frac{\vert P(k,z)^{\mathrm{emulated}}_\mathrm{lin} -P(k,z)^{\mathrm{true}}_\mathrm{lin} \vert} {P(k,z)^{\mathrm{true}}_\mathrm{lin} }$', fontsize=30)
        plt.xlabel(r'$k [h Mpc^{-1}]$', fontsize=30)
        # Save the plot
        plt.savefig(f'{self.save_dir}_{run}_test.png', dpi=200, bbox_inches='tight')
        plt.close()

    def run_training_cmb(self, tf, cosmopower_NN, cosmopower_PCAplusNN):
        """Train the neural network for the CMB power spectrum emulation."""
        raise NotImplementedError

    def run_testing_cmb(self, tf, cosmopower_NN, cosmopower_PCAplusNN):
        """Test the neural network for the CMB power spectrum emulation and plot the results."""
        raise NotImplementedError

    def config(self):
        self.TF_ENABLE_ONEDNN_OPTS = self.read_ini("TF_ENABLE_ONEDNN_OPTS", int, 0)
        self.CUDA_VISIBLE_DEVICES = self.read_ini("CUDA_VISIBLE_DEVICES", str, '0')
        self.np_seed = self.read_ini("np_seed", int, 0)
        self.tf_seed = self.read_ini("tf_seed", int, 2)

        os.environ['TF_ENABLE_ONEDNN_OPTS'] = str(self.TF_ENABLE_ONEDNN_OPTS)
        os.environ["CUDA_VISIBLE_DEVICES"]= str(self.CUDA_VISIBLE_DEVICES)
        import tensorflow as tf
        self.converged = False
        self.fatal_errors = self.read_ini("fatal_errors", bool, False)

        self.device = 'gpu:0' if tf.config.list_physical_devices('GPU') else 'cpu'

        self.save_dir = self.pipeline.save_name.replace('cosmopower_inputs', 'cosmopower_emulator')
        self.samples_name = self.pipeline.save_name
        self.tests_name =  f'{self.pipeline.save_name}_test'

    def execute(self):
        os.environ['TF_ENABLE_ONEDNN_OPTS'] = str(self.TF_ENABLE_ONEDNN_OPTS)
        os.environ["CUDA_VISIBLE_DEVICES"]= str(self.CUDA_VISIBLE_DEVICES)
        import tensorflow as tf    
        from cosmopower import cosmopower_NN, cosmopower_PCAplusNN
        # setting the seed for reproducibility
        np.random.seed(self.np_seed)
        tf.random.set_seed(self.tf_seed)


        # If our pipeline allows it we arrange it so that the
        # fast parameters change fastest in the sequence.
        # This is still not optimal for the multiprocessing case
        if self.pipeline.do_fast_slow:
            params = self.pipeline.slow_params + self.pipeline.fast_params
        else:
            params = self.pipeline.varied_params

        # We load the generated power spectra but interpolate them over specific k-range as defined here.
        # This makes sure our training has more features where power spectra change faster ...
        # How to deal with redshifts?? Ideally as done by Pierre, we train at specific z's that are also in latin hypercube
        # CAMB by default returns power spectra at an vector. 

        # Load your data
        files_train = glob.glob(f"{self.samples_name}_*.pkl")
        files_test  = glob.glob(f"{self.tests_name}_*.pkl")
        files_train = [f for f in files_train if f not in files_test]

        index_list_train, training_runs, params_dict_train, runs_dict = self.load_pickles(params, files_train)
        index_list_test, test_runs, params_dict_test, test_dict = self.load_pickles(params, files_test)

        assert set(training_runs) == set(test_runs), "The training and testing runs do not contain the same elements!"
            
        # Save fixed parameters to file that were used to generate power spectra.
        # This is useful later when we need to check if the input parameters are the same when using the emulator.
        fixed_params = {p.name: p.limits[0] for p in self.pipeline.fixed_params}
        with open(f'{self.save_dir}_fixed_params.pkl', 'wb') as f:
            pickle.dump(fixed_params, f)

        limits = {p.name: p.limits for p in params}
        with open(f'{self.save_dir}_param_limits.pkl', 'wb') as f:
            pickle.dump(limits, f)

        mean_params = [np.mean(params_dict_train[key]) for key in list(params_dict_train.keys())]
        mean_run = self.pipeline.run_results(mean_params)

        # outputs are either Plin, Pnl, CMB spectra
        for i, run in enumerate(training_runs):
            print(f'CosmoPower run {i+1}: training on {run}!')
            if run == 'cmb_cl':
                #run_training_cmb(tf, cosmopower_NN, cosmopower_PCAplusNN)
                #run_testing_cmb(tf, cosmopower_NN, cosmopower_PCAplusNN)
                print('Skipping CMB training, not yet implemented!')
                continue
            else:    
                self.run_training_pk(tf, cosmopower_NN, cosmopower_PCAplusNN, run, params_dict_train, runs_dict, mean_run)
                self.run_testing_pk(tf, cosmopower_NN, cosmopower_PCAplusNN, run, params_dict_test, test_dict, mean_run)

        self.converged = True


    def is_converged(self):
        return self.converged




import tempfile
import os
import cosmosis.getdist


def default_config():
    root = os.path.split(os.path.abspath(__file__))[0]

    params = {
        ('runtime', 'root'): root,
        ('runtime', 'verbosity'): "noisy",
        ("pipeline", "debug"): "F",
        ("pipeline", "modules"): "test1",
        ("pipeline", "extra_output"): "parameters/p3",
        ("test1", "file"): "example_module.py",
    }

    values = {
        ("parameters", "p1"): "-3.0 0.0 3.0",
        ("parameters", "p2"): "-3.0 0.0 3.0",
    }

    params = cosmosis.Inifile(None, override=params)
    values = cosmosis.Inifile(None, override=values)


    return params, values

def test_getdist_metropolis():
    with tempfile.TemporaryDirectory() as tempdir:
        params, values = default_config()
        params["runtime", "sampler"] = "metropolis"
        params.add_section("metropolis")
        params["metropolis", "samples"] = "1000"
        params["metropolis", "nsteps"] = "100"
        params.add_section("output")
        for i in [1, 2]:
            params["output", "filename"] = os.path.join(tempdir, f"chain_{i}.txt")
            cosmosis.run_cosmosis(params, values=values)

        chain_root = os.path.join(tempdir, "chain")
        cosmosis.getdist.cosmosis_to_getdist(chain_root, "metropolis", nchain=2, burn=0.1)

def test_getdist_emcee():
    with tempfile.TemporaryDirectory() as tempdir:
        params, values = default_config()
        params["runtime", "sampler"] = "emcee"
        params.add_section("emcee")
        params["emcee", "walkers"] = "16"
        params["emcee", "samples"] = "20"
        params["emcee", "nsteps"] = "5"
        params.add_section("output")
        params["output", "filename"] = os.path.join(tempdir, "chain.txt")
        cosmosis.run_cosmosis(params, values=values)

        chain_file = os.path.join(tempdir, "chain.txt")
        cosmosis.getdist.cosmosis_to_getdist(chain_file, "emcee", burn=0.1)

def test_getdist_nautilus():
    with tempfile.TemporaryDirectory() as tempdir:
        params, values = default_config()
        params["runtime", "sampler"] = "nautilus"
        params.add_section("nautilus")
        params["nautilus", "n_live"] = "100"
        params.add_section("output")
        params["output", "filename"] = os.path.join(tempdir, "chain.txt")
        cosmosis.run_cosmosis(params, values=values)

        chain_file = os.path.join(tempdir, "chain.txt")
        cosmosis.getdist.cosmosis_to_getdist(chain_file, "nautilus")



from ...runtime import logs
from numpy import pi, dot, exp, einsum
import numpy as np


class PopulationMonteCarlo(object):
	"""
	A Population Monte Carlo (PMC) sampler,
	which combines expectation-maximization and
	importance sampling
	
	This code follows the notation and methodolgy in
	http://arxiv.org/pdf/0903.0837v1.pdf


	"""
	def __init__(self, posterior, n, start, sigma, pool=None, student=False, nu=2.0):
		"""
		posterior: the posterior function
		n: number of components to use in the mixture
		start: estimated mean of the distribution
		sigma: estimated covariance matrix
		pool (optional): an MPI or multiprocessing worker pool

		"""
		self.posterior = posterior
		mu = np.random.multivariate_normal(start, sigma, size=n)

		if student:
			self.components = [StudentsTComponent(1.0/n, m, sigma, nu) for m in mu]
		else:
			self.components = [GaussianComponent(1.0/n, m, sigma) for m in mu]
		self.pool = pool

	def sample(self, n, update=True, do_kill=True):
		"Draw a sample from the Gaussian mixture and update the mixture"
		self.kill_count = n*1./len(self.components)/50.
		self.kill = [False for c in self.components]
		#draw sample from current mixture
		component_index, x = self.draw(n)

		#calculate likelihoods
		if self.pool is None:
			samples = list(map(self.posterior, x))
		else:
			samples = self.pool.map(self.posterior, x)

		post = np.array([s[0] for s in samples])
		extra = [s[1] for s in samples]
		post[np.isnan(post)] = -np.inf

		#update components
		log_weights = self.update_components(x, post, update, do_kill)

		return x, post, extra, component_index, log_weights

	def draw(self, n):
		"Draw a sample from the Gaussian mixture"
		A = [m.alpha for m in self.components]
		A = np.array(A)
		A/=A.sum()
		#Components to draw from
		N = np.arange(len(self.components))
		C = np.random.choice(N, size=n, replace=True, p=A)
		for i in N:
			count = np.sum(C==i)
			if count<self.kill_count:
				self.kill[i] = True
				logs.debug(f"Component {i} less than kill count ({count} < {self.kill_count})")
		x = np.array([self.components[c].sample() for c in C])
		return C, x

	def update_components(self, x, log_post, update, do_kill):
		"Equations 13-16 of arxiv.org 0903.0837v1"

		#x #n_sample*n_dim
		log_Aphi = np.array([np.log(m.alpha) + m.log_phi(x) for m in self.components]) #n-component * n_sample
		Aphi = np.array([m.alpha*m.phi(x) for m in self.components]) #n-component * n_sample
		post = np.exp(log_post)
		w = post/Aphi.sum(0) #n_sample
		logw = log_post - np.log(Aphi.sum(0))


		if not update:
			return logw

		w_norm = w/w.sum()  #n_sample

		logw_norm = np.log(w_norm)
		entropy =  -(w_norm*logw_norm).sum()
		perplexity = np.exp(entropy) / len(x)
		logs.debug(f"Perplexity = {perplexity}")


		Aphi[np.isnan(Aphi)] = 0.0
		w_norm[np.isnan(w_norm)] = 0.0
		A = [m.alpha for m in self.components]
		#rho_top = einsum('i,ij->ij', A, phi)  #n_component * n_sample
		rho_bottom = Aphi.sum(0) #n_sample
		rho = [rho_t/rho_bottom for rho_t in Aphi]



		for d,(m,rho_d) in enumerate(zip(self.components, rho)):
			try:
				m.update(w_norm, x, rho_d)
			except np.linalg.LinAlgError as error:
				logs.debug(f"Component not fitting the data very well {d} {error}")
				self.kill[d] = True

		if do_kill:
			self.components = [c for c,kill in zip(self.components,self.kill) if not kill]
		logs.debug("%d components remain" % len(self.components))
		return logw



class GaussianComponent(object):
	"""
	A single Gaussian component of the mixture model.

	"""
	def __init__(self,alpha, mu, sigma):
		self.set(alpha, mu, sigma)

	def set(self, alpha, mu, sigma):
		"Set the parameters of this distribution component"
		self.alpha = alpha
		self.mu = mu
		ndim = len(self.mu)
		self.sigma = sigma
		self.sigma_inv = np.linalg.inv(self.sigma)
		self.A = (2*pi)**(-ndim/2.0) * np.linalg.det(self.sigma)**-0.5
		self.logA = np.log(self.A)

	def update(self, w_norm, x, rho_d):
		"Update the parameters according to the samples and rho values"
		alpha = dot(w_norm, rho_d)  #scalar
		if not alpha>0:
			raise np.linalg.LinAlgError("alpha = %f"%alpha)
		mu = einsum('i,ij,i->j',w_norm, x, rho_d) / alpha  #scalar
		delta = x-mu  #n_sample * n_dim
		logs.debug(f"Updating to mu = {mu}")
		sigma = einsum('i,ij,ik,i->jk',w_norm, delta, delta, rho_d) / alpha  #n_dim * n_dim
		self.set(alpha, mu, sigma)

	def phi(self, x):
		"Evaluate the distribution"
		d = (x-self.mu) #n_sample * n_dim
		chi2 = einsum('ij,jk,ik->i',d,self.sigma_inv,d)
		#result size n_sample
		return self.A * exp(-0.5*chi2)

	def log_phi(self, x):
		"Evaluate the log distribution"
		d = (x-self.mu) #n_sample * n_dim
		chi2 = einsum('ij,jk,ik->i',d,self.sigma_inv,d)
		return self.logA - 0.5*chi2


	def sample(self):
		"Draw a sample from the distribution"
		return np.random.multivariate_normal(self.mu, self.sigma)




class StudentsTComponent(object):
	"""
	A single Students-t component of the mixture model.

	Implements (unnumbered) equations between A8
	and A9 in http://arxiv.org/pdf/0903.0837v1.pdf

	Not yet working
	"""
	def __init__(self,alpha, mu, sigma, nu):
		self.nu=nu
		self.ndim = len(mu)
		self.set(alpha, mu, sigma)

	def set(self, alpha, mu, sigma):
		"Set the parameters of this distribution component"
		from scipy.special import gamma
		self.alpha = alpha
		self.mu = mu
		p = self.ndim
		nu=self.nu
		self.sigma = sigma
		self.sigma_inv = np.linalg.inv(self.sigma)
		self.A = gamma((nu+p)/2.)/gamma(nu/2.) / (pi*nu)**(p/2.) * np.linalg.det(self.sigma)**-0.5

	def update(self, w_norm, x, rho_d):
		"Update the parameters according to the samples and rho values"
		alpha = dot(w_norm, rho_d)  #scalar
		if not alpha>0:
			raise np.linalg.LinAlgError("alpha = %f"%alpha)
		nu=self.nu
		p=self.ndim
		gamma = (nu+p)/(nu+self.chi2)  #size n_sampl
		mu = einsum('i,ij,i,i->j',w_norm, x, rho_d,gamma)  #scalar
		mu /= einsum('i,i,i',w_norm,rho_d,gamma)
		delta = x-mu  #n_sample * n_dim
		sigma = einsum('i,ij,ik,i,i->jk',w_norm, delta, delta, rho_d,gamma)  #n_dim * n_dim
		sigma/= einsum('i,i',w_norm,rho_d)
		self.set(alpha, mu, sigma)

	def phi(self, x):
		"Evaluate the distribution.  This is called tau in the paper"
		d = (x-self.mu) #n_sample * n_dim
		chi2 = einsum('ij,jk,ik->i',d,self.sigma_inv,d) #size n_sample
		#record this chi2 as we will need it later.
		#valid until x,mu,or sigma_d changes
		self.chi2=chi2
		#result size n_sample
		nu=self.nu
		p=self.ndim
		return self.A * (1.0+chi2/nu)**(-(nu+p)/2.0)

	def sample(self):
		"Draw a sample from the distribution"
		y = np.random.multivariate_normal(np.zeros_like(self.mu), self.sigma)
		z = np.random.chisquare(self.nu)
		return self.mu + y*(self.nu/z)**0.5



def test_like(p):
	x,y=p
	a = (x+y)/np.sqrt(2)
	b = (x-y)/np.sqrt(2)
	return 0.2*exp(-0.5*(a**2+b**2/25.0)) + 0.8*exp(-0.5*(a**2/25.0+b**2))


def test():
	pmc = PopulationMonteCarlo(test_like, 2, [0.0, 0.0], [1.0, 1.0])
	for i in range(50):
		s = pmc.sample(4000)
	import pylab
	x = s[:,0]
	y = s[:,1]
	# pylab.plot(x[:,0], x[:,1], '.')
	xmin=x.min()
	xmax=x.max()
	ymin=y.min()
	ymax=y.max()
	Y,X=np.mgrid[ymin:ymax:100j,xmin:xmax:100j]
	r = np.array(list(zip(X.flatten(),Y.flatten())))
	P = np.array([test_like((xi,yi) ) for (xi,yi) in r])
	P = P.reshape(X.shape)
	for c in pmc.components:
		pylab.contour(X, Y, c.phi(r).reshape(X.shape), 3)
	pylab.contourf(X,Y,P,50, cmap='Reds')
	pylab.plot(x, y, 'k,')
	pylab.show()


if __name__ == '__main__':
	test()
